/*===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 * 
 * IBM SDK, Java(tm) Technology Edition, v7
 * (C) Copyright IBM Corp. 2013, 2013. All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *===========================================================================
 */
/*
 * Copyright (c) 1995, 2011, Oracle and/or its affiliates. All rights reserved.
 * ORACLE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 */

/**
 * =======================================================================
 * Module Information:
 *
 * DESCRIPTION: This class implements RDMA socket by extending from
 * the PlainSockImpl.
 * 
 * This package private class provides RDMA_CAPABILITY.
 * =======================================================================
 */

package java.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileDescriptor;
import java.security.AccessController;
import sun.security.action.GetPropertyAction;

class RDMASocketImpl extends PlainSocketImpl {
	// associated input/output streams and abstract plain socket impl
	private RDMASocketInputStream socketInputStream = null;
	private RDMASocketOutputStream socketOutputStream = null;
	private AbstractPlainSocketImpl abstractPlainSocketImpl = null;
	// store the pointer of the native data structure related to this socket
	private long socketDataPtr = 0;
	private final SocketImpl oldImpl;
	private boolean created = false;
	private final int MAX_CONNECT_RETRY = 10;

	// tcp fallback support
	private static final boolean tcpFallback;
 	private static final int DFLT_TCP_FALLBACK_WAIT_TIME = 100; // milliseconds
	private static final int MIN_TCP_FALLBACK_WAIT_TIME = 20; // milliseconds
	private static final int MAX_TCP_FALLBACK_WAIT_TIME = 10000; // milliseconds
 	private static final int tcpFallbackWaitTime;
 	private boolean tcpFallbackInPlace = false;
 	private static final String connTimedOutStr = "Connection timed out";
 	private static final String acceptTimedOutStr = "Accept timed out";
		
	// native methods
	static native void initProtoRDMA();
	native long socketCreateRDMA(FileDescriptor fdObj) throws IOException;
	native void socketConnectRDMA(InetAddress address, int port, long socketDataPtr, int timeOut) throws IOException;
	native void socketBindRDMA(InetAddress address, int port, long socketDataPtr) throws IOException;
	native void socketListenRDMA(int count, long socketDataPtr) throws IOException;
	native long socketAcceptRDMA(SocketImpl acceptedSocketImpl, long serverSocketDataPtr, int timeOut) throws IOException;
	native void socketCloseRDMA(boolean useDeferredClose, long socketDataPtr) throws IOException;
	native void socketShutdownRDMA(int howto, long socketDataPtr) throws IOException;
	native void socketSendUrgentDataRDMA(int data, long socketDataPtr) throws IOException;
	native void socketSetOptionRDMA(int cmd, boolean on, Object value, long socketDataPtr) throws SocketException;
	native int socketGetOptionRDMA(int opt, Object iaContainerObj, long socketDataPtr) throws SocketException;
	native int socketAvailableRDMA(long socketDataPtr) throws IOException;
	
	// do native initialization only once
	static {
		boolean fallback = false;
 		String tcpFallbackPropStr = null;
 		String tcpFallbackWaitTimePropStr = null;
 		int fallbackWaitTime = DFLT_TCP_FALLBACK_WAIT_TIME;

		initProtoRDMA();

		tcpFallbackPropStr = AccessController.doPrivileged(new GetPropertyAction("com.ibm.net.rdma.tcpFallback"));
		if (tcpFallbackPropStr != null) {
 			fallback = Boolean.parseBoolean(tcpFallbackPropStr);
 		}
 		tcpFallback = fallback;
 
		tcpFallbackWaitTimePropStr = AccessController.doPrivileged(new GetPropertyAction("com.ibm.net.rdma.tcpFallbackWaitTime"));
		if (tcpFallbackWaitTimePropStr != null) {
			try {
				fallbackWaitTime = Integer.parseInt(tcpFallbackWaitTimePropStr);
				if (fallbackWaitTime < MIN_TCP_FALLBACK_WAIT_TIME) {
					System.out.format("Invalid tcp fallback wait time \'%d\' - defaulting to %d milliseconds%n",
							fallbackWaitTime, MIN_TCP_FALLBACK_WAIT_TIME);
					fallbackWaitTime = MIN_TCP_FALLBACK_WAIT_TIME;
				} else if (fallbackWaitTime > MAX_TCP_FALLBACK_WAIT_TIME) {
					System.out.format("Invalid tcp fallback wait time \'%d\' - defaulting to %d milliseconds%n",
							fallbackWaitTime, MAX_TCP_FALLBACK_WAIT_TIME);
					fallbackWaitTime = MAX_TCP_FALLBACK_WAIT_TIME;
				}
 			} catch (NumberFormatException e) {
				System.out.println("Invalid \"com.ibm.net.rdma.tcpFallbackWaitTime\" property value ("
						+e.getMessage()+"): defaulting to "+DFLT_TCP_FALLBACK_WAIT_TIME+" milliseconds");
				fallbackWaitTime = DFLT_TCP_FALLBACK_WAIT_TIME;
 			}
 		}
 		tcpFallbackWaitTime = fallbackWaitTime;
	}
	
	// helper method to create RDMA socket and set data pointer field
	private void socketCreatePrivate(FileDescriptor fdObj) throws IOException {
		socketDataPtr = socketCreateRDMA(fdObj);
		created = true;
	}
	
	/**
	 * Constructs a RDMASocketImpl instance
	 */
	RDMASocketImpl(SocketImpl oldImpl) {
		this.oldImpl = oldImpl;
	}
	
	/**
	 * Creates an RDMA stream socket
	 */
	@Override
	protected void create(boolean stream) throws IOException {
		if (!stream) {
			throw new UnsupportedOperationException("Must be a stream socket for RDMA!");
		}
		super.create(stream);
	}
	
	/**
	 * Store abstract plain socket impl that is used
	 */
	void setAbstractPlainSocketImpl(AbstractPlainSocketImpl impl) {
		abstractPlainSocketImpl = impl;
	}
	
	/**
	 * Return true if this RDMA socket is already closed or close is pending
	 */
	@Override
	public boolean isClosedOrPending() {
		if (super.isClosedOrPending()) {
			return true;
		}
		if (socketDataPtr == 0) {
			if (tcpFallbackInPlace == false) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Get input stream for this RDMA socket.
	 */
	@Override
	protected synchronized InputStream getInputStream() throws IOException {
		if (!tcpFallbackInPlace && socketDataPtr != 0) {
			if (isClosedOrPending()) {
				throw new IOException("Socket Closed");
			}
			if (inputShut()) {
				throw new IOException("Socket input is shutdown");
			}
			if (socketInputStream == null) {
				socketInputStream = new RDMASocketInputStream(this);
			}
			return socketInputStream;
		} else {
			return super.getInputStream();
		}
	}

	/**
	 * Get output stream for the RDMA socket.
	 */
	@Override
	protected synchronized OutputStream getOutputStream() throws IOException {
		if (!tcpFallbackInPlace && socketDataPtr != 0) {
			if (isClosedOrPending()) {
				throw new IOException("Socket Closed");
			}
			if (outputShut()) {
				throw new IOException("Socket output is shutdown");
			}
			if (socketOutputStream == null) {
				socketOutputStream = new RDMASocketOutputStream(this);
			}
			return socketOutputStream;
		} else {
			return super.getOutputStream();
		}
	}
	
	/**
	 * z/OS specific function hence null implementation to satisfy base class
	 */
	void processNRException(InetAddress address, int port, int backlogQ) {
		// do nothing
	}
	
	/**
	 * Create an RDMA stream socket
	 */
	@Override
	void socketCreate(boolean isServer) throws IOException {
		super.socketCreate(isServer);
		socketCreatePrivate(null);
	}
	
	/**
	 * Connect this RDMA socket to the specified port on the specified host
	 */
	@Override
	void socketConnect(InetAddress address, int port, int timeout) throws IOException {
		/* The socket connect can happen lately after creating a plain unconnected socket.
		 * In such a scenario, we need to ensure that RDMA portion of the socket is also
		 * created and linked to the plain socket before attempting the connect operation.
		 */
		if (!created && (oldImpl != null)) {
			socketCreatePrivate(oldImpl.getFileDescriptor());
		}
		int connRetryCnt = 0;

		do {
			try {
				connRetryCnt += 1;
				super.socketConnect(address, port, timeout);
			} catch (ConnectException ce) {
				if ((connRetryCnt < MAX_CONNECT_RETRY) && (ce.getMessage().equals(connTimedOutStr))) {
					continue;
				} else {
					throw new ConnectException(ce.getMessage());
				}
			} 
			break;
		} while (true);

		try {
			int waitTime = 0;
			if (timeout < 0) {
				System.out.format("Invalid timeout parameter \'%d\' - defaulting to zero (infinite timeout)%n", timeout);
				timeout = 0;
				waitTime = timeout;
			} else if (timeout > 0) {
				waitTime = timeout;
			} else {
				waitTime = 0;
			}
				
			socketConnectRDMA(address, port, socketDataPtr, waitTime);
		} catch (Exception e) {
			boolean connTimedOut = e.getMessage().equals(connTimedOutStr);

			socketCloseRDMA(false, socketDataPtr);
			socketDataPtr = 0;
			if (connTimedOut) {
				throw new SocketTimeoutException("RDMA Connect timed out");
			} else {
				throw new IOException("RDMA Connect failed");
			}
		}
	}
	
	/**
	 * Bind this RDMA socket to the specified IP address and port number
	 */
	@Override
	void socketBind(InetAddress address, int port) throws IOException {
		/* The socket bind can happen lately after a plain server socket creation.
		 * If that is the case, then we need to ensure that RDMA portion of the server
		 * socket is also created and linked to the plain server socket before attempting
		 * the bind operation.
		 */
		if (!created && (oldImpl != null)) {
			socketCreatePrivate(oldImpl.getFileDescriptor());
			created = true;
		}
                
		super.socketBind(address, port);
		socketBindRDMA(address, port, socketDataPtr);
	}
	
	/**
	 * Listen for incoming connection indications on this RDMA socket
	 * with the maximum queue length of 'count'
	 */
	@Override
	void socketListen(int count) throws IOException {
		super.socketListen(count);
		socketListenRDMA(count, socketDataPtr);
	}

	/**
 	 * Accept the incoming connections on the RDMA server socket
 	 */
	long socketPostAccept(SocketImpl acceptedSocketImpl, long serverSocketDataPtr) throws IOException {
		long clientDataSocketPtr = 0;

		try {
 			int waitTime = 0;
			int acceptTimeout = 0;

			acceptTimeout = ((PlainSocketImpl)this).getTimeout();
			if (acceptTimeout == 0 && (oldImpl != null)) {
				acceptTimeout = ((PlainSocketImpl)oldImpl).getTimeout();
			}

			if (tcpFallback) {
				waitTime = tcpFallbackWaitTime;
			} else if (acceptTimeout > 0) {
				waitTime = acceptTimeout;
 			} else {
				waitTime = 0;
			}

			clientDataSocketPtr = socketAcceptRDMA(acceptedSocketImpl, serverSocketDataPtr, waitTime);
		} catch (Exception e) {
			if (e.getMessage().equals(acceptTimedOutStr)) {
				if (tcpFallback) {
 					((RDMASocketImpl)acceptedSocketImpl).setTcpFallbackInPlace(true);
 					clientDataSocketPtr = 0;
				} else {
					throw new SocketTimeoutException("RDMA Accept timed out");
				}
			} else {
 				throw new IOException("RDMA Accept failed");
			}
		}
		return clientDataSocketPtr;
    }
	
	/**
	 * Close this RDMA socket
	 */
	@Override
	void socketClose0(boolean useDeferredClose) throws IOException {
		// close the RDMA socket followed by the plain socket
		if (socketDataPtr != 0) {
			socketCloseRDMA(useDeferredClose, socketDataPtr);
		}
		super.socketClose0(useDeferredClose);
	}
	
	/**
	 * Shutdown this RDMA connection
	 */
	@Override
	void socketShutdown(int howto) throws IOException {
		// shutdown the RDMA socket followed by the plain socket
		if (!tcpFallbackInPlace && socketDataPtr != 0) {
			socketShutdownRDMA(howto, socketDataPtr);
		}
		super.socketShutdown(howto);
	}
	
	/**
	 * Send urgent data on this RDMA socket connection
	 */
	@Override
	void socketSendUrgentData(int data) throws IOException {
		if (!tcpFallbackInPlace && socketDataPtr != 0) {
			socketSendUrgentDataRDMA(data, socketDataPtr);
		} else {
			super.socketSendUrgentData(data);
		}
	}

	/**
	 * Returns the local port to which this socket is bound.
	 */
	@Override
	protected int getLocalPort() {
		if (abstractPlainSocketImpl != null) {
			return(abstractPlainSocketImpl.getLocalPort());
		} else {
			return(super.getLocalPort());
		}
	}
	
	/**
	 * Returns the remote port to which this socket is connected.
	 */
	@Override
	protected int getPort() {
		if (abstractPlainSocketImpl != null) {
			return(abstractPlainSocketImpl.getPort());
		} else {
			return(super.getPort());
		}
	}

	/**
 	 * Returns the address to which the socket is connected.
 	 */
	@Override
	protected InetAddress getInetAddress() {
		if (abstractPlainSocketImpl != null) {
			return (abstractPlainSocketImpl.getInetAddress());
 		} else {
			return (super.getInetAddress());
		}
                
	}

	/**
 	 * socketSetOption()
 	 */
	@Override
	void socketSetOption(int cmd, boolean on, Object value) throws SocketException {
		super.socketSetOption(cmd, on, value);

		if (!tcpFallbackInPlace && socketDataPtr != 0) {
			switch (cmd) {
			case SO_REUSEADDR:
			case SO_RCVBUF:
			case SO_SNDBUF:
			case SO_LINGER:
			case SO_KEEPALIVE:
			case SO_OOBINLINE:
				socketSetOptionRDMA(cmd, on, value, socketDataPtr);
				break;
			default:
				// other socket related options are not supported on RDMA socket
				break;
			}
		}
		return;
	}
        
 	/**
 	 * socketGetOption()
 	 */
	@Override
	int socketGetOption(int opt, Object iaContainerObj) throws SocketException {
		int optVal = -1;

		if (!tcpFallbackInPlace && socketDataPtr != 0) {
			switch (opt) {
			case SO_REUSEADDR:
			case SO_RCVBUF:
			case SO_SNDBUF:
			case SO_LINGER:
			case SO_KEEPALIVE:
			case SO_OOBINLINE:
				optVal = socketGetOptionRDMA(opt, iaContainerObj, socketDataPtr);
				break;
			default:
				// other socket related options are not supported on RDMA socket
				optVal = -1;
				break;
			}
		}

		/* fallback to TCP for unsupported options */
		if (optVal < 0) {
			optVal = super.socketGetOption(opt, iaContainerObj);
		}
		return optVal;
	}

        /**
         * Method to get available number of bytes to read.
         * It overrides InputStream.available
         */
        @Override
        int socketAvailable() throws IOException {
               return socketAvailableRDMA(socketDataPtr);
        }

	/**
	 * Getter method for the RDMA socket pointer
	 */
	long getSocketDataPtr() {
		return socketDataPtr;
	}
	
	/**
	 * Setter method for the RDMA socket pointer
	 */
	void setSocketDataPtr(long socketDataPtr) {
		this.socketDataPtr = socketDataPtr;
	}

	/**
 	 * Setter method for the tcp fallback status field
 	 */
	void setTcpFallbackInPlace(boolean fallback) {
		this.tcpFallbackInPlace = fallback;
	}
}
